/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 1991-1997 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.	IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "agent.h"
#include "packet.h"
#include "ip.h"
#include "timer-handler.h"
#include "random.h"
#include "tfrc-formula.h"

#define SMALLFLOAT 0.0000001

/* receiver response */ 

#define DECREASE 1
#define NORMAL 2
#define INCREASE 3

#define MAXSEQ 1073741824   /* Number that curseq_ is set to for */
                                /* "infinite send" (2^30)            */


/* modes of rate change */
#define SLOW_START 1
#define CONG_AVOID 2
#define RATE_DECREASE	 3
#define OUT_OF_SLOW_START 4 

#define RDO_SIZE 41 //Number of items in RDO list

//Sequences, 4cif format, rate (kb/s) psnr curve, x264 encoder, QPI=QPP=[10,50], baseline profile
/*
const double rate_psnr[41][2] = { {128.28, 25.222}, {138.29, 25.739}, {146.24, 26.149}, {157.18, 26.603}, {170.85, 27.065},
		{184.64, 27.533}, {200.77, 27.973}, {224.25, 28.508}, {245.94, 28.956}, {273.33, 29.454},
		{308.83, 30.014}, {351.26, 30.576}, {393.54, 31.064}, {460.23, 31.717}, {520.06, 32.219},
		{606.05, 32.796}, {714.00, 33.435}, {841.34, 34.032}, {995.15, 34.587}, {1232.82, 35.293},
		{1444.28, 35.881}, {1751.58, 36.484}, {2203.21, 37.182}, {2736.46, 37.785}, {3355.30, 38.406},
		{4306.73, 39.187}, {5246.88, 39.791}, {6393.86, 40.457}, {8051.15, 41.273}, {9633.88, 41.995},
		{11509.41, 42.757}, {13636.42, 43.589}, {15803.41, 44.375}, {18004.48, 45.155}, {20715.53, 46.058},
		{23093.56, 46.877}, {26078.55, 47.786}, {29376.81, 48.771}, {32124.38, 49.557}, {34974.83, 50.363}, {39061.70, 51.413}
};
*/

/*
//CIF format
const double rate_psnr_akiyo[41][2] = { { 14.57 , 28.429 }, { 17.36 , 29.091 }, { 18.61 , 29.592 }, { 19.39 , 30.120 }, { 22.10 , 30.639 },
	{ 22.77 , 31.257 }, { 24.14 , 31.737 }, { 25.94 , 32.289 }, { 27.33 , 32.682 }, { 29.81 , 33.250 },
	{ 32.27 , 33.792 }, { 35.85 , 34.394 }, { 38.48 , 34.908 }, { 42.81 , 35.614 }, { 46.64 , 36.121 },
	{ 52.18 , 36.723 }, { 58.08 , 37.332 }, { 64.72 , 37.857 }, { 72.82 , 38.413 }, { 84.93 , 39.092 },
	{ 93.97 , 39.628 }, { 105.97 , 40.183 }, { 123.86 , 40.844 }, { 142.29 , 41.401 }, { 161.92 , 41.939 },
	{ 191.10 , 42.675 }, { 215.07 , 43.112 }, { 247.35 , 43.662 }, { 291.94 , 44.276 }, { 336.71 , 44.769 },
	{ 387.84 , 45.284 }, { 469.26 , 45.869 }, { 540.99 , 46.334 }, { 636.51 , 46.817 }, { 715.31 , 47.355 },
	{ 818.41 , 47.978 }, { 969.51 , 48.576 }, { 1160.35 , 49.263 }, { 1354.10 , 49.802 }, { 1592.40 , 50.419 },{ 1953.48 , 51.169 }
};

const double rate_psnr_coastguard[41][2] = { { 36.05 , 24.365 }, { 39.96 , 24.813 }, { 43.26 , 25.181 }, { 47.48 , 25.566 }, { 53.19 , 25.974 },
	{ 60.19 , 26.407 }, { 68.23 , 26.823 }, { 80.63 , 27.309 }, { 92.36 , 27.729 }, { 109.33 , 28.192 },
	{ 132.33 , 28.736 }, { 160.52 , 29.271 }, { 191.70 , 29.769 }, { 239.81 , 30.422 }, { 283.64 , 30.944 },
	{ 344.29 , 31.562 }, { 420.69 , 32.226 }, { 505.48 , 32.862 }, { 599.07 , 33.492 }, { 728.98 , 34.232 },
	{ 844.37 , 34.897 }, { 988.81 , 35.582 }, { 1168.06 , 36.367 }, { 1354.31 , 37.063 }, { 1549.71 , 37.773 },
	{ 1809.71 , 38.677 }, { 2035.50 , 39.365 }, { 2302.92 , 40.098 }, { 2652.33 , 41.008 }, { 2982.50 , 41.777 },
	{ 3345.94 , 42.593 }, { 3798.57 , 43.443 }, { 4216.13 , 44.242 }, { 4666.13 , 45.006 }, { 5200.59 , 45.891 },
	{ 5679.97 , 46.753 }, { 6274.02 , 47.623 }, { 6963.37 , 48.536 }, { 7565.87 , 49.267 }, { 8233.41 , 50.030 }, { 9238.70 , 51.038 }
};

const double rate_psnr_container[41][2] = { { 22.10 , 25.207 }, { 24.88 , 25.769 }, { 26.68 , 26.239 }, { 28.87 , 26.732 }, { 31.16 , 27.218 },
	{ 33.09 , 27.774 }, { 35.59 , 28.319 }, { 38.93 , 28.852 }, { 41.81 , 29.328 }, { 45.57 , 29.886 },
	{ 50.26 , 30.453 }, { 55.92 , 31.020 }, { 61.48 , 31.516 }, { 70.00 , 32.185 }, { 77.33 , 32.656 },
	{ 88.77 , 33.228 }, { 103.00 , 33.845 }, { 119.62 , 34.417 }, { 138.52 , 34.958 }, { 169.19 , 35.616 },
	{ 193.52 , 36.181 }, { 227.75 , 36.738 }, { 276.44 , 37.414 }, { 330.13 , 38.008 }, { 390.22 , 38.629 },
	{ 478.77 , 39.459 }, { 557.06 , 40.019 }, { 660.89 , 40.672 }, { 805.22 , 41.506 }, { 944.11 , 42.173 },
	{ 1099.26 , 42.919 }, { 1341.14 , 43.714 }, { 1550.87 , 44.455 }, { 1795.63 , 45.188 }, { 2083.54 , 45.967 },
	{ 2382.51 , 46.749 }, { 2798.04 , 47.532 }, { 3286.01 , 48.358 }, { 3774.38 , 49.090 }, { 4320.36 , 49.810 }, { 5195.98 , 50.733 }
};

const double rate_psnr_foreman[41][2] = { { 46.97 , 25.847 }, { 51.56 , 26.472 }, { 55.48 , 26.957 }, { 61.22 , 27.491 }, { 65.74 , 28.010 },
	{ 70.96 , 28.455 }, { 77.14 , 28.922 }, { 84.08 , 29.465 }, { 91.06 , 29.929 }, { 99.87 , 30.386 },
	{ 109.41 , 30.923 }, { 122.28 , 31.438 }, { 134.45 , 31.920 }, { 153.85 , 32.537 }, { 170.02 , 33.010 },
	{ 193.45 , 33.554 }, { 220.75 , 34.156 }, { 251.38 , 34.704 }, { 288.33 , 35.237 }, { 341.52 , 35.898 },
	{ 387.09 , 36.482 }, { 445.31 , 37.030 }, { 526.66 , 37.712 }, { 615.80 , 38.297 }, { 713.04 , 38.905 },
	{ 847.77 , 39.654 }, { 974.35 , 40.213 }, { 1129.60 , 40.828 }, { 1344.03 , 41.594 }, { 1558.20 , 42.233 },
	{ 1808.63 , 42.888 }, { 2147.34 , 43.625 }, { 2479.65 , 44.322 }, { 2862.11 , 45.021 }, { 3345.65 , 45.812 },
	{ 3799.56 , 46.601 }, { 4393.09 , 47.408 }, { 5074.42 , 48.284 }, { 5716.36 , 49.038 }, { 6415.02 , 49.798 }, { 7453.86 , 50.802 }
};

const double rate_psnr_hall[41][2] = { { 20.03 , 25.561 }, { 22.66 , 26.193 }, { 24.98 , 26.788 }, { 27.72 , 27.389 }, { 30.87 , 28.036 },
	{ 33.82 , 28.715 }, { 38.42 , 29.316 }, { 42.72 , 29.964 }, { 46.59 , 30.485 }, { 53.01 , 31.122 },
	{ 58.87 , 31.814 }, { 67.67 , 32.440 }, { 74.40 , 33.005 }, { 87.95 , 33.695 }, { 96.80 , 34.190 },
	{ 112.96 , 34.788 }, { 131.87 , 35.406 }, { 148.13 , 35.954 }, { 183.71 , 36.513 }, { 220.08 , 37.140 },
	{ 256.35 , 37.635 }, { 292.32 , 38.087 }, { 366.48 , 38.663 }, { 451.77 , 39.142 }, { 537.60 , 39.606 },
	{ 665.77 , 40.179 }, { 791.94 , 40.597 }, { 948.17 , 41.057 }, { 1218.63 , 41.630 }, { 1505.40 , 42.160 },
	{ 1894.72 , 42.791 }, { 2370.70 , 43.470 }, { 2986.48 , 44.217 }, { 3617.52 , 44.981 }, { 4409.20 , 45.908 },
	{ 5132.77 , 46.860 }, { 5920.97 , 47.773 }, { 6829.51 , 48.741 }, { 7544.95 , 49.508 }, { 8340.36 , 50.331 }, { 9384.30 , 51.359 }
};

const double rate_psnr_mobile[41][2] = { { 91.52 , 20.624 }, { 97.77 , 21.191 }, { 104.12 , 21.643 }, { 118.43 , 22.175 }, { 129.42 , 22.707 },
	{ 142.29 , 23.240 }, { 165.37 , 23.792 }, { 185.62 , 24.398 }, { 204.44 , 24.909 }, { 236.71 , 25.525 },
	{ 270.60 , 26.195 }, { 316.44 , 26.854 }, { 357.20 , 27.462 }, { 427.22 , 28.225 }, { 482.46 , 28.796 },
	{ 571.71 , 29.493 }, { 684.29 , 30.252 }, { 791.99 , 30.879 }, { 939.80 , 31.609 }, { 1126.04 , 32.420 },
	{ 1299.68 , 33.142 }, { 1487.84 , 33.784 }, { 1758.80 , 34.659 }, { 2028.95 , 35.448 }, { 2317.06 , 36.257 },
	{ 2689.83 , 37.223 }, { 3009.91 , 37.995 }, { 3381.66 , 38.839 }, { 3855.42 , 39.851 }, { 4290.19 , 40.717 },
	{ 4758.10 , 41.616 }, { 5321.70 , 42.588 }, { 5815.41 , 43.457 }, { 6357.41 , 44.329 }, { 7034.97 , 45.343 },
	{ 7619.27 , 46.220 }, { 8310.40 , 47.180 }, { 9088.96 , 48.189 }, { 9756.57 , 49.004 }, { 10476.53 , 49.848 },{ 11510.97 , 50.944 }
};

const double rate_psnr_mother[41][2] = { { 12.70 , 28.661 }, { 14.63 , 29.385 }, { 15.87 , 29.990 }, { 17.92 , 30.379 }, { 20.50 , 30.785 },
	{ 22.90 , 31.208 }, { 25.86 , 31.582 }, { 29.45 , 32.051 }, { 32.92 , 32.540 }, { 36.55 , 33.227 },
	{ 41.23 , 33.808 }, { 46.29 , 34.336 }, { 51.01 , 34.716 }, { 58.17 , 35.347 }, { 63.80 , 35.784 },
	{ 72.57 , 36.258 }, { 81.24 , 36.859 }, { 91.61 , 37.421 }, { 103.82 , 37.901 }, { 120.74 , 38.544 },
	{ 133.41 , 39.119 }, { 150.67 , 39.703 }, { 175.79 , 40.327 }, { 206.80 , 40.884 }, { 237.47 , 41.417 },
	{ 272.78 , 42.192 }, { 312.93 , 42.652 }, { 361.91 , 43.161 }, { 439.76 , 43.848 }, { 516.03 , 44.350 },
	{ 600.54 , 44.863 }, { 745.37 , 45.467 }, { 878.76 , 45.947 }, { 1049.00 , 46.443 }, { 1274.19 , 46.962 },
	{ 1526.78 , 47.539 }, { 1949.73 , 48.138 }, { 2509.51 , 48.822 }, { 3088.10 , 49.535 }, { 3729.30 , 50.230 }, { 4680.69 , 51.155 }
};

const double rate_psnr_news[41][2] = { { 26.95 , 25.884 }, { 30.03 , 26.474 }, { 32.50 , 26.948 }, { 36.48 , 27.494 }, { 39.99 , 28.031 },
	{ 43.60 , 28.568 }, { 48.50 , 29.151 }, { 53.70 , 29.806 }, { 58.09 , 30.288 }, { 64.97 , 30.925 },
	{ 71.61 , 31.585 }, { 80.32 , 32.225 }, { 87.77 , 32.762 }, { 100.29 , 33.543 }, { 109.34 , 34.089 },
	{ 123.04 , 34.743 }, { 138.73 , 35.428 }, { 154.12 , 36.038 }, { 172.67 , 36.664 }, { 197.96 , 37.369 },
	{ 218.33 , 37.981 }, { 243.62 , 38.566 }, { 277.84 , 39.292 }, { 314.31 , 39.875 }, { 352.32 , 40.508 },
	{ 404.93 , 41.297 }, { 449.47 , 41.836 }, { 503.71 , 42.438 }, { 580.76 , 43.143 }, { 652.38 , 43.720 },
	{ 734.81 , 44.270 }, { 863.15 , 44.961 }, { 978.52 , 45.496 }, { 1125.22 , 46.046 }, { 1284.48 , 46.686 },
	{ 1465.27 , 47.383 }, { 1716.85 , 48.051 }, { 2030.94 , 48.804 }, { 2344.23 , 49.435 }, { 2708.57 , 50.123 }, { 3266.73 , 50.965 }
};

const double rate_psnr_silent[41][2] = { { 24.16 , 26.136 }, { 26.95 , 26.631 }, { 28.76 , 27.131 }, { 32.67 , 27.552 }, { 36.52 , 27.897 },
	{ 40.83 , 28.319 }, { 46.74 , 28.796 }, { 52.90 , 29.332 }, { 58.28 , 29.683 }, { 65.95 , 30.168 },
	{ 74.40 , 30.664 }, { 84.82 , 31.152 }, { 93.97 , 31.544 }, { 108.87 , 32.079 }, { 121.20 , 32.535 },
	{ 138.06 , 33.037 }, { 157.59 , 33.590 }, { 178.80 , 34.122 }, { 203.22 , 34.650 }, { 236.99 , 35.293 },
	{ 264.52 , 35.863 }, { 298.49 , 36.400 }, { 345.50 , 37.067 }, { 394.46 , 37.693 }, { 448.05 , 38.351 },
	{ 519.99 , 39.189 }, { 584.80 , 39.760 }, { 664.38 , 40.425 }, { 772.43 , 41.271 }, { 879.42 , 41.947 },
	{ 999.88 , 42.665 }, { 1178.73 , 43.522 }, { 1330.47 , 44.211 }, { 1522.57 , 44.938 }, { 1745.69 , 45.763 },
	{ 1965.27 , 46.498 }, { 2309.07 , 47.300 }, { 2744.85 , 48.146 }, { 3185.68 , 48.861 }, { 3732.52 , 49.579 },{ 4617.94 , 50.467 }
};
*/


const double rate_psnr_city[41][2] = { {128.28, 25.222}, {138.29, 25.739}, {146.24, 26.149}, {157.18, 26.603}, {170.85, 27.065},
		{184.64, 27.533}, {200.77, 27.973}, {224.25, 28.508}, {245.94, 28.956}, {273.33, 29.454},
		{308.83, 30.014}, {351.26, 30.576}, {393.54, 31.064}, {460.23, 31.717}, {520.06, 32.219},
		{606.05, 32.796}, {714.00, 33.435}, {841.34, 34.032}, {995.15, 34.587}, {1232.82, 35.293},
		{1444.28, 35.881}, {1751.58, 36.484}, {2203.21, 37.182}, {2736.46, 37.785}, {3355.30, 38.406},
		{4306.73, 39.187}, {5246.88, 39.791}, {6393.86, 40.457}, {8051.15, 41.273}, {9633.88, 41.995},
		{11509.41, 42.757}, {13636.42, 43.589}, {15803.41, 44.375}, {18004.48, 45.155}, {20715.53, 46.058},
		{23093.56, 46.877}, {26078.55, 47.786}, {29376.81, 48.771}, {32124.38, 49.557}, {34974.83, 50.363}, {39061.70, 51.413}
};

const double rate_psnr_crew[41][2] = { {266.30, 28.792}, {274.37, 29.212}, {284.23, 29.587}, {310.93, 30.028}, {328.12, 30.436},
		{351.63, 30.885}, {397.51, 31.347}, {433.90, 31.800}, {468.11, 32.223}, {527.34, 32.678},
		{585.12, 33.161}, {668.14, 33.653}, {732.19, 34.074}, {852.77, 34.649}, {944.33, 35.075},
		{1085.90, 35.580}, {1242.74, 36.091}, {1405.40, 36.568}, {1612.94, 37.018}, {1892.43, 37.541},
		{2144.00, 38.038}, {2430.70, 38.458}, {2843.81, 38.972}, {3335.98, 39.415}, {3862.94, 39.877},
		{4658.91, 40.478}, {5448.00, 40.938}, {6407.05, 41.439}, {7840.78, 42.120}, {9288.67, 42.746},
		{11019.80, 43.448}, {13268.91, 44.197}, {15440.38, 45.042}, {17651.27, 45.803}, {20424.39, 46.739},
		{22992.73, 47.618}, {25706.45, 48.445}, {29048.37, 49.415}, {31439.09, 50.084}, {34300.08, 50.848},{37934.94, 51.766}
};

const double rate_psnr_harbour[41][2] = { {137.97, 24.048}, {160.04, 24.559}, {180.23, 24.957}, {208.21, 25.446}, {242.16, 25.956},
		{284.48, 26.481}, {330.54, 26.957}, {401.71, 27.600}, {468.06, 28.097}, {557.32, 28.671},
		{676.90, 29.312}, {817.82, 29.955}, {959.20, 30.525}, {1180.22, 31.275}, {1373.94, 31.860},
		{1634.33, 32.522}, {1950.59, 33.244}, {2299.47, 33.894}, {2685.36, 34.532}, {3210.53, 35.249},
		{3697.13, 35.907}, {4292.24, 36.536}, {5039.57, 37.267}, {5874.85, 37.901}, {6734.00, 38.555},
		{7956.53, 39.334}, {9040.18, 39.976}, {10284.42, 40.663}, {12005.28, 41.488}, {13603.44, 42.235},
		{15406.57, 43.011}, {17587.62, 43.851}, {19638.74, 44.615}, {21759.59, 45.380}, {24388.28, 46.253},
		{26860.32, 47.080}, {29680.22, 47.987}, {32895.39, 48.958}, {35442.45, 49.697}, {38298.55, 50.496}, {42243.96, 51.514}
};

const double rate_psnr_soccer[41][2] = { {181.41, 28.115}, {194.93, 28.536}, {210.03, 28.847}, {228.14, 29.214}, {248.41, 29.574},
		{272.82, 29.929}, {303.82, 30.296}, {340.53, 30.713}, {375.61, 31.057}, {423.55, 31.437},
		{481.34, 31.890}, {552.09, 32.315}, {619.74, 32.735}, {725.21, 33.287}, {816.61, 33.735},
		{941.14, 34.250}, {1089.74, 34.816}, {1253.50, 35.367}, {1441.22, 35.887}, {1708.55, 36.529},
		{1930.70, 37.122}, {2223.21, 37.689}, {2623.90, 38.349}, {3075.23, 38.929}, {3563.82, 39.528},
		{4264.75, 40.304}, {4950.11, 40.860}, {5780.12, 41.476}, {6964.69, 42.240}, {8127.95, 42.895},
		{9493.24, 43.589}, {11215.89, 44.328}, {12917.01, 45.056}, {14708.65, 45.764}, {16934.90, 46.603},
		{18966.31, 47.410}, {21471.29, 48.231}, {24404.62, 49.150}, {26872.59, 49.900}, {29565.64, 50.678}, {33255.46, 51.664}
};


//Mean Rate/PSNR between 4 sequences (City, Crew, Harbour, Soccer)
/*
const double rate_psnr_mean[111][2] = { {338.73, 29.00}, {354.24, 29.20}, {371.12, 29.40}, {389.45, 29.60}, {410.05, 29.80},
		{431.24, 30.00}, {453.84, 30.20}, {476.44, 30.40}, {500.48, 30.60}, {526.87, 30.80},
		{555.01, 31.00}, {585.53, 31.20}, {617.53, 31.40}, {649.58, 31.60}, {681.99, 31.80},
		{718.17, 32.00}, {756.17, 32.20}, {797.58, 32.40}, {840.13, 32.60}, {884.23, 32.80},
		{930.11, 33.00}, {976.48, 33.20}, {1029.03, 33.40}, {1084.78, 33.60}, {1140.93, 33.80},
		{1199.93, 34.00}, {1265.03, 34.20}, {1332.48, 34.40}, {1402.55, 34.60}, {1479.82, 34.80},
		{1558.74, 35.00}, {1639.85, 35.20}, {1723.59, 35.40}, {1810.74, 35.60}, {1899.10, 35.80},
		{1998.28, 36.00}, {2108.16, 36.20}, {2218.82, 36.40}, {2334.90, 36.60}, {2460.16, 36.80},
		{2585.42, 37.00}, {2717.84, 37.20}, {2875.50, 37.40}, {3037.68, 37.60}, {3201.83, 37.80},
		{3373.10, 38.00}, {3551.42, 38.20}, {3733.60, 38.40}, {3940.08, 38.60}, {4158.55, 38.80},
		{4379.84, 39.00}, {4616.65, 39.20}, {4871.23, 39.40}, {5132.74, 39.60}, {5397.53, 39.80},
		{5676.62, 40.00}, {5964.67, 40.20}, {6260.62, 40.40}, {6588.10, 40.60}, {6937.03, 40.80},
		{7297.37, 41.00}, {7666.30, 41.20}, {8040.35, 41.40}, {8432.82, 41.60}, {8832.19, 41.80},
		{9231.89, 42.00}, {9648.87, 42.20}, {10088.70, 42.40}, {10532.40, 42.60}, {10979.20, 42.80},
		{11440.30, 43.00}, {11918.90, 43.20}, {12398.20, 43.40}, {12899.40, 43.60}, {13433.80, 43.80},
		{13971.50, 44.00}, {14509.90, 44.20}, {15027.50, 44.40}, {15548.20, 44.60}, {16072.90, 44.80},
		{16597.90, 45.00}, {17145.20, 45.20}, {17707.00, 45.40}, {18279.40, 45.60}, {18853.00, 45.80},
		{19434.40, 46.00}, {20012.40, 46.20}, {20588.10, 46.40}, {21163.60, 46.60}, {21731.70, 46.80},
		{22310.00, 47.00}, {22899.20, 47.20}, {23490.80, 47.40}, {24107.70, 47.60}, {24742.60, 47.80},
		{25382.80, 48.00}, {26032.40, 48.20}, {26687.90, 48.40}, {27350.90, 48.60}, {28016.80, 48.80},
		{28690.50, 49.00}, {29370.70, 49.20}, {30054.60, 49.40}, {30744.90, 49.60}, {31440.50, 49.80},
		{32143.60, 50.00}, {32855.80, 50.20}, {33575.00, 50.40}, {34316.50, 50.60}, {35073.80, 50.80}, {35844.70, 51.00}
};
*/

//Mean Rate/PSNR between 2 sequences (Crew, Harbour)

const double rate_psnr_mean[111][2] = { {444.50, 29.00}, {465.07, 29.20}, {487.75, 29.40}, {512.52, 29.60}, {540.49, 29.80},
		{569.11, 30.00}, {598.38, 30.20}, {627.40, 30.40}, {659.00, 30.60}, {693.71, 30.80},
		{731.12, 31.00}, {770.51, 31.20}, {811.69, 31.40}, {852.84, 31.60}, {893.99, 31.80},
		{939.54, 32.00}, {986.96, 32.20}, {1038.75, 32.40}, {1092.84, 32.60}, {1149.02, 32.80},
		{1204.78, 33.00}, {1261.51, 33.20}, {1329.89, 33.40}, {1400.43, 33.60}, {1469.76, 33.80},
		{1542.26, 34.00}, {1621.58, 34.20}, {1703.04, 34.40}, {1788.83, 34.60}, {1883.44, 34.80},
		{1978.18, 35.00}, {2077.01, 35.20}, {2178.82, 35.40}, {2281.07, 35.60}, {2385.71, 35.80},
		{2499.96, 36.00}, {2627.13, 36.20}, {2755.84, 36.40}, {2888.91, 36.60}, {3037.27, 36.80},
		{3185.62, 37.00}, {3340.64, 37.20}, {3515.94, 37.40}, {3700.29, 37.60}, {3882.66, 37.80},
		{4064.84, 38.00}, {4261.11, 38.20}, {4460.74, 38.40}, {4674.72, 38.60}, {4912.03, 38.80},
		{5153.64, 39.00}, {5421.68, 39.20}, {5693.62, 39.40}, {5976.26, 39.60}, {6259.11, 39.80},
		{6554.74, 40.00}, {6868.30, 40.20}, {7181.85, 40.40}, {7519.25, 40.60}, {7890.73, 40.80},
		{8277.03, 41.00}, {8677.04, 41.20}, {9077.06, 41.40}, {9495.45, 41.60}, {9919.93, 41.80},
		{10344.40, 42.00}, {10777.20, 42.20}, {11237.60, 42.40}, {11701.30, 42.60}, {12169.10, 42.80},
		{12648.00, 43.00}, {13152.80, 43.20}, {13659.00, 43.40}, {14206.10, 43.60}, {14766.00, 43.80},
		{15332.50, 44.00}, {15900.60, 44.20}, {16426.10, 44.40}, {16951.50, 44.60}, {17485.10, 44.80},
		{18019.30, 45.00}, {18580.00, 45.20}, {19150.10, 45.40}, {19741.80, 45.60}, {20333.40, 45.80},
		{20930.70, 46.00}, {21528.10, 46.20}, {22123.90, 46.40}, {22719.00, 46.60}, {23313.00, 46.80},
		{23904.10, 47.00}, {24502.40, 47.20}, {25105.50, 47.40}, {25708.60, 47.60}, {26344.40, 47.80},
		{26984.70, 48.00}, {27644.00, 48.20}, {28303.30, 48.40}, {28975.20, 48.60}, {29650.90, 48.80},
		{30329.40, 49.00}, {31018.60, 49.20}, {31707.70, 49.40}, {32408.80, 49.60}, {33117.40, 49.80},
		{33832.20, 50.00}, {34557.00, 50.20}, {35288.90, 50.40}, {36036.50, 50.60}, {36798.50, 50.80}, {37576.90, 51.00}
};


//Mean Rate/PSNR between 2 sequences (Crew, Soccer)
/*
const double rate_psnr_mean[111][2] = { {243.94, 29.00}, {250.79, 29.20}, {258.96, 29.40}, {267.61, 29.60}, {280.54, 29.80},
		{294.03, 30.00}, {306.94, 30.20}, {319.79, 30.40}, {333.64, 30.60}, {348.29, 30.80},
		{366.42, 31.00}, {388.28, 31.20}, {410.32, 31.40}, {431.09, 31.60}, {451.88, 31.80},
		{474.86, 32.00}, {499.60, 32.20}, {528.47, 32.40}, {557.59, 32.60}, {587.05, 32.80},
		{618.12, 33.00}, {650.14, 33.20}, {686.86, 33.40}, {724.13, 33.60}, {761.42, 33.80},
		{800.81, 34.00}, {843.83, 34.20}, {890.54, 34.40}, {937.76, 34.60}, {985.38, 34.80},
		{1036.32, 35.00}, {1091.62, 35.20}, {1150.43, 35.40}, {1214.83, 35.60}, {1281.62, 35.80},
		{1351.54, 36.00}, {1425.73, 36.20}, {1501.47, 36.40}, {1577.65, 36.60}, {1661.24, 36.80},
		{1744.82, 37.00}, {1840.57, 37.20}, {1945.60, 37.40}, {2049.80, 37.60}, {2157.06, 37.80},
		{2268.39, 38.00}, {2394.01, 38.20}, {2527.35, 38.40}, {2682.02, 38.60}, {2840.21, 38.80},
		{3004.03, 39.00}, {3196.70, 39.20}, {3389.36, 39.40}, {3587.92, 39.60}, {3792.31, 39.80},
		{4008.00, 40.00}, {4230.77, 40.20}, {4469.35, 40.40}, {4748.90, 40.60}, {5043.71, 40.80},
		{5352.72, 41.00}, {5678.89, 41.20}, {6005.05, 41.40}, {6359.19, 41.60}, {6724.78, 41.80},
		{7090.36, 42.00}, {7464.24, 42.20}, {7868.62, 42.40}, {8277.51, 42.60}, {8690.53, 42.80},
		{9124.77, 43.00}, {9568.10, 43.20}, {10011.40, 43.40}, {10497.60, 43.60}, {11030.90, 43.80},
		{11564.30, 44.00}, {12097.10, 44.20}, {12587.40, 44.40}, {13078.00, 44.60}, {13568.70, 44.80},
		{14059.30, 45.00}, {14590.40, 45.20}, {15134.00, 45.40}, {15677.60, 45.60}, {16223.40, 45.80},
		{16784.90, 46.00}, {17346.50, 46.20}, {17908.10, 46.40}, {18469.80, 46.60}, {19016.70, 46.80},
		{19560.60, 47.00}, {20104.50, 47.20}, {20648.40, 47.40}, {21243.10, 47.60}, {21873.10, 47.80},
		{22506.40, 48.00}, {23139.60, 48.20}, {23784.80, 48.40}, {24444.80, 48.60}, {25108.50, 48.80},
		{25772.20, 49.00}, {26438.40, 49.20}, {27112.00, 49.40}, {27797.40, 49.60}, {28483.90, 49.80},
		{29178.80, 50.00}, {29892.30, 50.20}, {30612.90, 50.40}, {31333.50, 50.60}, {32071.30, 50.80}, {32836.30, 51.00}
};
*/
//Mean Rate/PSNR between 2 sequences (City, Crew)
/*
const double rate_psnr_mean[111][2] = { {259.33, 29.00}, {266.75, 29.20}, {274.84, 29.40}, {283.80, 29.60}, {296.19, 29.80},
		{308.59, 30.00}, {320.52, 30.20}, {332.29, 30.40}, {345.02, 30.60}, {358.92, 30.80},
		{375.52, 31.00}, {395.17, 31.20}, {414.81, 31.40}, {433.06, 31.60}, {452.01, 31.80},
		{472.02, 32.00}, {492.02, 32.20}, {519.09, 32.40}, {547.01, 32.60}, {574.33, 32.80},
		{603.19, 33.00}, {633.00, 33.20}, {666.77, 33.40}, {704.20, 33.60}, {741.18, 33.80},
		{777.72, 34.00}, {823.26, 34.20}, {871.94, 34.40}, {921.01, 34.60}, {976.04, 34.80},
		{1031.20, 35.00}, {1090.44, 35.20}, {1153.37, 35.40}, {1217.63, 35.60}, {1284.29, 35.80},
		{1359.87, 36.00}, {1443.38, 36.20}, {1528.44, 36.40}, {1623.40, 36.60}, {1734.22, 36.80},
		{1845.04, 37.00}, {1964.66, 37.20}, {2106.54, 37.40}, {2247.58, 37.60}, {2387.47, 37.80},
		{2537.74, 38.00}, {2702.30, 38.20}, {2870.21, 38.40}, {3068.23, 38.60}, {3270.42, 38.80},
		{3476.92, 39.00}, {3712.04, 39.20}, {3978.79, 39.40}, {4248.29, 39.60}, {4518.75, 39.80},
		{4816.33, 40.00}, {5120.99, 40.20}, {5425.65, 40.40}, {5776.24, 40.60}, {6150.88, 40.80},
		{6531.69, 41.00}, {6926.21, 41.20}, {7330.97, 41.40}, {7757.00, 41.60}, {8186.74, 41.80},
		{8617.16, 42.00}, {9082.13, 42.20}, {9559.56, 42.40}, {10037.00, 42.60}, {10520.60, 42.80},
		{11022.80, 43.00}, {11525.10, 43.20}, {12027.30, 43.40}, {12571.50, 43.60}, {13147.50, 43.80},
		{13723.40, 44.00}, {14298.80, 44.20}, {14832.30, 44.40}, {15371.40, 44.60}, {15910.60, 44.80},
		{16449.80, 45.00}, {17019.50, 45.20}, {17610.20, 45.40}, {18201.00, 45.60}, {18791.70, 45.80},
		{19388.20, 46.00}, {19977.70, 46.20}, {20564.30, 46.40}, {21150.90, 46.60}, {21736.30, 46.80},
		{22342.20, 47.00}, {22962.80, 47.20}, {23583.40, 47.40}, {24203.90, 47.60}, {24857.70, 47.80},
		{25520.70, 48.00}, {26183.70, 48.20}, {26846.70, 48.40}, {27522.30, 48.60}, {28203.90, 48.80},
		{28897.90, 49.00}, {29592.00, 49.20}, {30286.10, 49.40}, {30993.00, 49.60}, {31704.00, 49.80},
		{32415.00, 50.00}, {33135.90, 50.20}, {33870.60, 50.40}, {34634.30, 50.60}, {35398.00, 50.80},{36178.10, 51.00}
};
*/

struct hdr_tfrc {

	// RFC 3448 specifies that the data packet include the sequence
	// number, timestamp, and RTT estimate.
	int seqno;		//data sequence number
	double timestamp; 	//time this message was sent
	double rtt;	 	//RTT estimate of sender
	// "rate" is used by one of the experimental algorithms, RBPH.
	double rate;		//sender's current rate
	// In a real implementation, tzero, psize, and fsize
	// would not be in the packet header.
	// They are here for convenience.
	double tzero;	 	//RTO in Umass eqn
	int psize;		//Packet size.
	int fsize;		//The default large packet size for VoIP.  
 	// UrgentFlag is used to request that a loss report be sent
	//  immediately.
	int UrgentFlag;		//Urgent Flag
	// "round_id" is used by PreciseLoss_, a variant for more
	//  precise loss events that is on by default.
	int round_id ; 		//round id.

	static int offset_;	// offset for this header
	inline static int& offset() { 
		return offset_; 
	}
	inline static hdr_tfrc* access(const Packet* p) {
		return (hdr_tfrc*) p->access(offset_);
	}

};

struct hdr_tfrc_ack {

	// RFC 3448 specifies that feedback packets must include the
	// timestamp_echo, timestamp_offset, flost, and 
	// rate_since_last_report. 
	double timestamp_offset;	//offset since we received data packet
	double timestamp_echo;		//timestamp from the last data packet
	double flost;		//frequency of loss indications
	double rate_since_last_report;	//what it says ...
	// Used in optional variants:
	int losses;		// number of losses in last RTT
	double NumFeedback_;	//number of times/RTT feedback is to be sent 
	// Used for statistics-reporting only:
	double true_loss;	// true loss event rate.  
	// Not used:
	int seqno;	 	// not sure yet
	//Tuan
	short congestionEvent; //1: if receiver state that there is congestion, 0 otherwise
	//Tuan
	double timestamp;		//time this nack was sent


	static int offset_;		 // offset for this header
	inline static int& offset() { 
		return offset_; 
	}
	inline static hdr_tfrc_ack* access(const Packet* p) {
		return (hdr_tfrc_ack*) p->access(offset_);
	}
};

class TfrcPSNRAgent; 

class TfrcPSNRSendTimer : public TimerHandler {
public:
		TfrcPSNRSendTimer(TfrcPSNRAgent *a) : TimerHandler() { a_ = a; }
		virtual void expire(Event *e);
protected:
		TfrcPSNRAgent *a_;
};	

class TfrcPSNRNoFeedbackTimer : public TimerHandler {
public:
		TfrcPSNRNoFeedbackTimer(TfrcPSNRAgent *a) : TimerHandler() { a_ = a; }
		virtual void expire(Event *e);
protected:
		TfrcPSNRAgent *a_;
}; 

class TfrcPSNRAgent : public Agent {
		friend class TfrcPSNRSendTimer;
	friend class TfrcPSNRNoFeedbackTimer;
public:
	TfrcPSNRAgent();
	void recv(Packet*, Handler*);
	void sendpkt();
	void nextpkt();
	int command(int argc, const char*const* argv);
	void start();
	void stop();
	void update_rtt(double tao, double now); 
	void increase_rate(double p);
	void increase_rate_virtual(double p);
	void increase_rate_virtual_new(double p);
	void increase_rate_psnr(double p);
	double rate_to_psnr(double array[][2], int size, double r);
	double psnr_to_rate(double array[][2], int size, double psnr);
	void decrease_rate();
	void decrease_rate_virtual();
	void decrease_rate_virtual_new();
	void decrease_rate_psnr();
	double rfc3390(int size);
	double initial_rate();
	void slowstart();
	void reduce_rate_on_no_feedback();
	void reduce_rate_on_no_feedback_psnr();
	void advanceby(int delta); 
	void sendmsg(int nbytes, const char *flags = 0);
protected:
	TfrcPSNRSendTimer send_timer_;
	TfrcPSNRNoFeedbackTimer NoFeedbacktimer_;

	/* "accurate" estimates for formula */
	double rtt_; /*EWMA version*/
	double rttcur_; /*Instantaneous version*/
	double rttvar_;
	double tzero_;
	double sqrtrtt_; /*The mean of the sqrt of the RTT*/

	int ca_; //Enable Sqrt(RTT) based congestion avoidance mode

	/* TCP variables for tracking RTT */
	int t_srtt_; 
	int t_rtt_;
	int t_rttvar_;
	int rttvar_exp_;
	double t_rtxcur_;
	double tcp_tick_;
	int T_SRTT_BITS; 
	int T_RTTVAR_BITS;
	int srtt_init_; 
	int rttvar_init_;
	double rtxcur_init_;
	/* End of TCP variables for tracking RTT */

	// Dynamic state:
	int maxseq_;            // max seq produced by the app so far
	int seqno_;             // next seq to be sent 
	int psize_;
	double rate_;		// send rate
	double oldrate_;	// allows rate to be changed gradually
	double delta_;		// allows rate to be changed gradually
	int rate_change_; 	// slow start, cong avoid, decrease ...
	double last_change_;	// time last change in rate was made
	double rcvrate  ; 	// TCP friendly rate based on current RTT 
				//  and recever-provded loss estimate
	double maxrate_;	// prevents sending at more than 2 times the 
				//  rate at which the receiver is _receving_ 
	double ss_maxrate_;	// max rate for during slowstart
	TracedInt ndatapack_;	// number of packets sent
	TracedInt ndatabytes_;	// number of bytes sent
	TracedDouble true_loss_rate_;	// true loss event rate,
	int active_;		// have we shut down? 
	int round_id ;		// round id
	int first_pkt_rcvd ;	// first ack received yet?
	double last_pkt_time_;	// time the last data packet was sent
	int maxqueue_;		// max queue from application
	int UrgentFlag;		// urgent flag
	int all_idle_;		// has the sender been idle since the
				//  nofeedback timer was set?
        double lastlimited_;	// time sender was last datalimited.
	// End of dynamic state.

	// Parameters:
	int InitRate_;		// initial send rate
	double df_;		// decay factor for accurate RTT estimate
	double ssmult_;		// during slow start, increase rate by this 
				//  factor every rtt
	int bval_;		// value of B for the formula
	double overhead_;	// if > 0, dither outgoing packets 
	int ecn_ ;		// Set to 1 for an ECN-capable connection.
	double minrto_ ;	// for experimental purposes, for a minimum
				//  RTO value (for use in the TCP-friendly
				//  equation).
	double rate_init_;	// Static value for initial rate, in 
				//   packets per RTT.
					// for statistics only
	int SndrType_;          // 0 -> infinite sender, 1 -> need FTP
 	int printStatus_;	// to print status reports
        // End of parameters:

	/* Variants in the TFRC algorithms.  */
//	int standard_;		// 1 for RFC 3448 algorithms.
                                // 2 for RFC 4342 algorithms.
                                // 3 for RFC 3448bis algorithms.
        int rate_init_option_;  /* 1 for using static rate_init_ */
                                /* 2 for using RFC 3390 */
	int slow_increase_;	// To use slow increases in the rate during
				//  slow-start.
	int datalimited_;	// to send immediately when a new packet
				//   arrives after a data-limited period
	int oldCode_;		// set to 1 not to use "datalimited_"
				//   parameter.
	int heavyrounds_;	// the number of RTTs so far when the
				//  sending rate > 2 * receiving rate
	int maxHeavyRounds_;	// the number of allowed rounds for
				//  sending rate > 2 * receiving rate
        int useHeaders_;        /* boolean: Add DCCP/IP header sizes */  
	int idleFix_;		// 1 for fix for receive rate limits
				//   when sender has been idle
	/* End of variants.  */

	/* Responses to heavy congestion. */
	int conservative_;	// set to 1 for an experimental, conservative 
				//   response to heavy congestion
	double scmult_;         // self clocking parameter for conservative_
	/* End of responses to heavy congestion.  */

	/* VoIP mode, for using small packets. */
	int voip_;		// 1 for voip mode.
	int voip_max_pkt_rate_ ;	// Max pkt rate in pps, for voip mode.
	int fsize_;		// Default size for large TCP packets 
				//  (e.g., 1460 bytes).
        int headersize_;	// Size for packet headers.
	/* end of VoIP mode. */

        int tfrc_psnr_debug_;

        double ai_psnr_; //Addictive increase PSNR in delta
        double md_psnr_; //Multiplicative decrease PSNR in percentage

        int algo_choice_; //1: pure AIMD, 2: virtual rate/PSNR curve
        int aimd_strategy_;

        double min_psnr_; //Min PSNR (equivalent to bit rate) the video must sent
        double max_psnr_; //Max PSNR (equivalent to bit rate) the video can be sent

        int vid_profile_; //1: city, 2: crew, 3: harbour, 4: soccer
        double rate_psnr[41][2];
        double rate_psnr_virtual[111][2];
        int decreased_last_time; //Indicate that we have decreased the rate from last report
        int decreased_count;
        int count;

        int previous_congestion_event;
        int loss_flag;

        double virtual_rate_;
};
